import inspect
import sys

import pytest
from dagster import check
from dagster.core.test_utils import ExplodingRunLauncher

from .graphql_context_test_suite import GraphQLContextVariant, manage_graphql_context


@pytest.mark.graphql_context_variants
@pytest.mark.parametrize("variant", GraphQLContextVariant.all_readonly_variants())
def test_readonly_variants(variant):
    # Note: This mark is only run in the `tox_postgres.ini` file and not the default `tox.ini`
    # file because 3 of the variants have a postgres dependency
    assert isinstance(variant, GraphQLContextVariant)
    with manage_graphql_context(variant) as context:
        assert isinstance(context.instance.run_launcher, ExplodingRunLauncher)


def get_all_static_functions(klass):
    check.invariant(sys.version_info >= (3,))

    def _yield_all():
        for attr_name in dir(klass):
            attr = inspect.getattr_static(klass, attr_name)
            if isinstance(attr, staticmethod):
                # the actual function is on the __func__ property
                yield attr.__func__

    return list(_yield_all())


def test_get_all_static_members():
    class Bar:
        class_var = "foo"

        @staticmethod
        def static_one():
            pass

        @staticmethod
        def static_two():
            pass

        @classmethod
        def classthing(cls):
            pass

    assert set(get_all_static_functions(Bar)) == {Bar.static_one, Bar.static_two}


def test_all_variants_in_variants_function():
    """
    This grabs all pre-defined variants on GraphQLContextVariant (defined as static methods that
    return a single ContextVariant) and tests two things:
    1) They all contain a unique test_id
    2) That the all_variants() static method returns *all* of them
    """

    variant_test_ids_declared_on_class = set()
    for static_function in get_all_static_functions(GraphQLContextVariant):
        maybe_variant = static_function()
        if isinstance(maybe_variant, GraphQLContextVariant):
            assert maybe_variant.test_id
            assert maybe_variant.test_id not in variant_test_ids_declared_on_class
            variant_test_ids_declared_on_class.add(maybe_variant.test_id)

    test_ids_returned_by_all_variants = {
        var.test_id for var in GraphQLContextVariant.all_variants()
    }

    assert test_ids_returned_by_all_variants == variant_test_ids_declared_on_class


def test_readonly_marks_filter():
    readonly_test_ids = {
        var.test_id
        for var in [
            GraphQLContextVariant.readonly_in_memory_instance_lazy_repository(),
            GraphQLContextVariant.readonly_in_memory_instance_multi_location(),
            GraphQLContextVariant.readonly_in_memory_instance_managed_grpc_env(),
            GraphQLContextVariant.readonly_sqlite_instance_lazy_repository(),
            GraphQLContextVariant.readonly_sqlite_instance_multi_location(),
            GraphQLContextVariant.readonly_sqlite_instance_managed_grpc_env(),
            GraphQLContextVariant.readonly_sqlite_instance_deployed_grpc_env(),
            GraphQLContextVariant.readonly_postgres_instance_lazy_repository(),
            GraphQLContextVariant.readonly_postgres_instance_multi_location(),
            GraphQLContextVariant.readonly_postgres_instance_managed_grpc_env(),
        ]
    }

    assert {
        var.test_id for var in GraphQLContextVariant.all_readonly_variants()
    } == readonly_test_ids



import graphene
from dagster.core.definitions.events import AssetKey

from ...implementation.execution import (
    cancel_partition_backfill,
    create_and_launch_partition_backfill,
    delete_pipeline_run,
    launch_pipeline_execution,
    launch_pipeline_reexecution,
    terminate_pipeline_execution,
    wipe_asset,
)
from ...implementation.external import (
    fetch_repository_locations,
    get_full_external_pipeline_or_raise,
)
from ...implementation.utils import (
    ExecutionMetadata,
    ExecutionParams,
    UserFacingGraphQLError,
    capture_error,
    pipeline_selector_from_graphql,
)
from ..asset_key import GrapheneAssetKey
from ..backfill import GrapheneCancelBackfillResult, GrapheneLaunchBackfillResult
from ..errors import (
    GrapheneAssetNotFoundError,
    GrapheneConflictingExecutionParamsError,
    GraphenePipelineRunNotFoundError,
    GraphenePresetNotFoundError,
    GraphenePythonError,
    GrapheneReloadNotSupported,
    GrapheneRepositoryLocationNotFound,
)
from ..external import (
    GrapheneRepositoryLocation,
    GrapheneRepositoryLocationConnection,
    GrapheneRepositoryLocationLoadFailure,
)
from ..inputs import GrapheneAssetKeyInput, GrapheneExecutionParams, GrapheneLaunchBackfillParams
from ..pipelines.pipeline import GraphenePipelineRun
from ..runs import GrapheneLaunchPipelineExecutionResult, GrapheneLaunchPipelineReexecutionResult
from ..schedules import (
    GrapheneReconcileSchedulerStateMutation,
    GrapheneStartScheduleMutation,
    GrapheneStopRunningScheduleMutation,
)
from ..sensors import GrapheneStartSensorMutation, GrapheneStopSensorMutation


def create_execution_params(graphene_info, graphql_execution_params):
    preset_name = graphql_execution_params.get("preset")
    selector = pipeline_selector_from_graphql(graphql_execution_params["selector"])
    if preset_name:
        if graphql_execution_params.get("runConfigData"):
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(conflicting_param="runConfigData")
            )

        if graphql_execution_params.get("mode"):
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(conflicting_param="mode")
            )

        if selector.solid_selection:
            raise UserFacingGraphQLError(
                GrapheneConflictingExecutionParamsError(
                    conflicting_param="selector.solid_selection"
                )
            )

        external_pipeline = get_full_external_pipeline_or_raise(graphene_info, selector)

        if not external_pipeline.has_preset(preset_name):
            raise UserFacingGraphQLError(
                GraphenePresetNotFoundError(preset=preset_name, selector=selector)
            )

        preset = external_pipeline.get_preset(preset_name)

        return ExecutionParams(
            selector=selector.with_solid_selection(preset.solid_selection),
            run_config=preset.run_config,
            mode=preset.mode,
            execution_metadata=create_execution_metadata(
                graphql_execution_params.get("executionMetadata")
            ),
            step_keys=graphql_execution_params.get("stepKeys"),
        )

    return execution_params_from_graphql(graphql_execution_params)


def execution_params_from_graphql(graphql_execution_params):
    return ExecutionParams(
        selector=pipeline_selector_from_graphql(graphql_execution_params.get("selector")),
        run_config=graphql_execution_params.get("runConfigData") or {},
        mode=graphql_execution_params.get("mode"),
        execution_metadata=create_execution_metadata(
            graphql_execution_params.get("executionMetadata")
        ),
        step_keys=graphql_execution_params.get("stepKeys"),
    )


def create_execution_metadata(graphql_execution_metadata):
    return (
        ExecutionMetadata(
            run_id=graphql_execution_metadata.get("runId"),
            tags={t["key"]: t["value"] for t in graphql_execution_metadata.get("tags", [])},
            root_run_id=graphql_execution_metadata.get("rootRunId"),
            parent_run_id=graphql_execution_metadata.get("parentRunId"),
        )
        if graphql_execution_metadata
        else ExecutionMetadata(run_id=None, tags={})
    )


class GrapheneDeletePipelineRunSuccess(graphene.ObjectType):
    runId = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeletePipelineRunSuccess"


class GrapheneDeletePipelineRunResult(graphene.Union):
    class Meta:
        types = (
            GrapheneDeletePipelineRunSuccess,
            GraphenePythonError,
            GraphenePipelineRunNotFoundError,
        )
        name = "DeletePipelineRunResult"


class GrapheneDeleteRunMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneDeletePipelineRunResult)

    class Arguments:
        runId = graphene.NonNull(graphene.String)

    class Meta:
        name = "DeleteRunMutation"

    def mutate(self, graphene_info, **kwargs):
        run_id = kwargs["runId"]
        return delete_pipeline_run(graphene_info, run_id)


class GrapheneTerminatePipelineExecutionSuccess(graphene.ObjectType):
    run = graphene.Field(graphene.NonNull(GraphenePipelineRun))

    class Meta:
        name = "TerminatePipelineExecutionSuccess"


class GrapheneTerminatePipelineExecutionFailure(graphene.ObjectType):
    run = graphene.NonNull(GraphenePipelineRun)
    message = graphene.NonNull(graphene.String)

    class Meta:
        name = "TerminatePipelineExecutionFailure"


class GrapheneTerminatePipelineExecutionResult(graphene.Union):
    class Meta:
        types = (
            GrapheneTerminatePipelineExecutionSuccess,
            GrapheneTerminatePipelineExecutionFailure,
            GraphenePipelineRunNotFoundError,
            GraphenePythonError,
        )
        name = "TerminatePipelineExecutionResult"


@capture_error
def create_execution_params_and_launch_pipeline_exec(graphene_info, execution_params_dict):
    # refactored into a helper function here in order to wrap with @capture_error,
    # because create_execution_params may raise
    return launch_pipeline_execution(
        graphene_info,
        execution_params=create_execution_params(graphene_info, execution_params_dict),
    )


class GrapheneLaunchPipelineExecutionMutation(graphene.Mutation):

    Output = graphene.NonNull(GrapheneLaunchPipelineExecutionResult)

    class Arguments:
        executionParams = graphene.NonNull(GrapheneExecutionParams)

    class Meta:
        description = "Launch a pipeline run via the run launcher configured on the instance."
        name = "LaunchPipelineExecutionMutation"

    def mutate(self, graphene_info, **kwargs):
        return create_execution_params_and_launch_pipeline_exec(
            graphene_info, kwargs["executionParams"]
        )


class GrapheneLaunchBackfillMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneLaunchBackfillResult)

    class Arguments:
        backfillParams = graphene.NonNull(GrapheneLaunchBackfillParams)

    class Meta:
        description = "Launches a set of partition backfill runs via the run launcher configured on the instance."
        name = "LaunchBackfillMutation"

    def mutate(self, graphene_info, **kwargs):
        return create_and_launch_partition_backfill(graphene_info, kwargs["backfillParams"])


class GrapheneCancelBackfillMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneCancelBackfillResult)

    class Arguments:
        backfillId = graphene.NonNull(graphene.String)

    class Meta:
        description = "Marks a partition backfill as canceled."
        name = "CancelBackfillMutation"

    def mutate(self, graphene_info, **kwargs):
        return cancel_partition_backfill(graphene_info, kwargs["backfillId"])


@capture_error
def create_execution_params_and_launch_pipeline_reexec(graphene_info, execution_params_dict):
    # refactored into a helper function here in order to wrap with @capture_error,
    # because create_execution_params may raise
    return launch_pipeline_reexecution(
        graphene_info,
        execution_params=create_execution_params(graphene_info, execution_params_dict),
    )


class GrapheneLaunchPipelineReexecutionMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneLaunchPipelineReexecutionResult)

    class Arguments:
        executionParams = graphene.NonNull(GrapheneExecutionParams)

    class Meta:
        description = "Re-launch a pipeline run via the run launcher configured on the instance"
        name = "LaunchPipelineReexecutionMutation"

    def mutate(self, graphene_info, **kwargs):
        return create_execution_params_and_launch_pipeline_reexec(
            graphene_info,
            execution_params_dict=kwargs["executionParams"],
        )


class GrapheneTerminatePipelinePolicy(graphene.Enum):
    # Default behavior: Only mark as canceled if the termination is successful, and after all
    # resources peforming the execution have been shut down.
    SAFE_TERMINATE = "SAFE_TERMINATE"

    # Immediately mark the pipelie as canceled, whether or not the termination was successful.
    # No guarantee that the execution has actually stopped.
    MARK_AS_CANCELED_IMMEDIATELY = "MARK_AS_CANCELED_IMMEDIATELY"

    class Meta:
        name = "TerminatePipelinePolicy"


class GrapheneTerminatePipelineExecutionMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneTerminatePipelineExecutionResult)

    class Arguments:
        runId = graphene.NonNull(graphene.String)
        terminatePolicy = graphene.Argument(GrapheneTerminatePipelinePolicy)

    class Meta:
        name = "TerminatePipelineExecutionMutation"

    def mutate(self, graphene_info, **kwargs):
        return terminate_pipeline_execution(
            graphene_info,
            kwargs["runId"],
            kwargs.get("terminatePolicy", GrapheneTerminatePipelinePolicy.SAFE_TERMINATE),
        )


class GrapheneReloadRepositoryLocationMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneRepositoryLocation,
            GrapheneReloadNotSupported,
            GrapheneRepositoryLocationNotFound,
            GrapheneRepositoryLocationLoadFailure,
        )
        name = "ReloadRepositoryLocationMutationResult"


class GrapheneReloadRepositoryLocationMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneReloadRepositoryLocationMutationResult)

    class Arguments:
        repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        name = "ReloadRepositoryLocationMutation"

    def mutate(self, graphene_info, **kwargs):
        location_name = kwargs["repositoryLocationName"]

        if not graphene_info.context.has_repository_location(
            location_name
        ) and not graphene_info.context.has_repository_location_error(location_name):
            return GrapheneRepositoryLocationNotFound(location_name)

        if not graphene_info.context.is_reload_supported(location_name):
            return GrapheneReloadNotSupported(location_name)

        # The current workspace context is a WorkspaceRequestContext, which contains a reference to the
        # repository locations that were present in the root WorkspaceProcessContext the start of the
        # request. Reloading a repository location modifies the WorkspaceWorkspaceProcessContext, rendeirng
        # our current WorkspaceRequestContext outdated. Therefore, `reload_repository_location` returns
        # an updated WorkspaceRequestContext for us to use.
        new_context = graphene_info.context.reload_repository_location(location_name)

        if new_context.has_repository_location(location_name):
            return GrapheneRepositoryLocation(new_context.get_repository_location(location_name))
        else:
            return GrapheneRepositoryLocationLoadFailure(
                location_name, new_context.get_repository_location_error(location_name)
            )


class GrapheneReloadWorkspaceMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneRepositoryLocationConnection,
            GraphenePythonError,
        )
        name = "ReloadWorkspaceMutationResult"


class GrapheneReloadWorkspaceMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneReloadWorkspaceMutationResult)

    class Meta:
        name = "ReloadWorkspaceMutation"

    def mutate(self, graphene_info, **_kwargs):
        new_context = graphene_info.context.reload_workspace()
        return fetch_repository_locations(new_context)


class GrapheneAssetWipeSuccess(graphene.ObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKey)

    class Meta:
        name = "AssetWipeSuccess"


class GrapheneAssetWipeMutationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetNotFoundError,
            GraphenePythonError,
            GrapheneAssetWipeSuccess,
        )
        name = "AssetWipeMutationResult"


class GrapheneAssetWipeMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneAssetWipeMutationResult)

    class Arguments:
        assetKey = graphene.Argument(graphene.NonNull(GrapheneAssetKeyInput))

    class Meta:
        name = "AssetWipeMutation"

    def mutate(self, graphene_info, **kwargs):
        return wipe_asset(graphene_info, AssetKey.from_graphql_input(kwargs["assetKey"]))


class GrapheneMutation(graphene.ObjectType):
    launch_pipeline_execution = GrapheneLaunchPipelineExecutionMutation.Field()
    launch_pipeline_reexecution = GrapheneLaunchPipelineReexecutionMutation.Field()
    reconcile_scheduler_state = GrapheneReconcileSchedulerStateMutation.Field()
    start_schedule = GrapheneStartScheduleMutation.Field()
    stop_running_schedule = GrapheneStopRunningScheduleMutation.Field()
    start_sensor = GrapheneStartSensorMutation.Field()
    stop_sensor = GrapheneStopSensorMutation.Field()
    terminate_pipeline_execution = GrapheneTerminatePipelineExecutionMutation.Field()
    delete_pipeline_run = GrapheneDeleteRunMutation.Field()
    reload_repository_location = GrapheneReloadRepositoryLocationMutation.Field()
    reload_workspace = GrapheneReloadWorkspaceMutation.Field()
    wipe_asset = GrapheneAssetWipeMutation.Field()
    launch_partition_backfill = GrapheneLaunchBackfillMutation.Field()
    cancel_partition_backfill = GrapheneCancelBackfillMutation.Field()

    class Meta:
        name = "Mutation"



import graphene
from dagster import check
from dagster.core.definitions.events import AssetKey
from dagster.utils.error import SerializableErrorInfo

from ..implementation.utils import PipelineSelector
from .util import non_null_list


class GrapheneError(graphene.Interface):
    message = graphene.String(required=True)

    class Meta:
        name = "Error"


class GraphenePythonError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PythonError"

    className = graphene.Field(graphene.String)
    stack = non_null_list(graphene.String)
    cause = graphene.Field(lambda: GraphenePythonError)

    def __init__(self, error_info):
        super().__init__()
        check.inst_param(error_info, "error_info", SerializableErrorInfo)
        self.message = error_info.message
        self.stack = error_info.stack
        self.cause = error_info.cause
        self.className = error_info.cls_name


class GrapheneSchedulerNotDefinedError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "SchedulerNotDefinedError"

    def __init__(self):
        super().__init__()
        self.message = "Scheduler is not defined for the currently loaded repository."


class GraphenePipelineSnapshotNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineSnapshotNotFoundError"

    snapshot_id = graphene.NonNull(graphene.String)

    def __init__(self, snapshot_id):
        super().__init__()
        self.snapshot_id = check.str_param(snapshot_id, "snapshot_id")
        self.message = f"Pipeline snapshot {snapshot_id} is not present in the current instance."


class GrapheneReloadNotSupported(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ReloadNotSupported"

    def __init__(self, location_name):
        super().__init__()
        self.message = f"Location {location_name} does not support reloading."


class GrapheneRepositoryLocationNotFound(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "RepositoryLocationNotFound"

    def __init__(self, location_name):
        super().__init__()
        self.message = f"Location {location_name} does not exist."


class GraphenePipelineNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineNotFoundError"

    pipeline_name = graphene.NonNull(graphene.String)
    repository_name = graphene.NonNull(graphene.String)
    repository_location_name = graphene.NonNull(graphene.String)

    def __init__(self, selector):
        super().__init__()
        check.inst_param(selector, "selector", PipelineSelector)
        self.pipeline_name = selector.pipeline_name
        self.repository_name = selector.repository_name
        self.repository_location_name = selector.location_name
        self.message = (
            "Could not find Pipeline "
            f"{selector.location_name}.{selector.repository_name}.{selector.pipeline_name}"
        )


class GraphenePipelineRunNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineRunNotFoundError"

    run_id = graphene.NonNull(graphene.String)

    def __init__(self, run_id):
        super().__init__()
        self.run_id = check.str_param(run_id, "run_id")
        self.message = f"Pipeline run {run_id} could not be found."


class GrapheneInvalidPipelineRunsFilterError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "InvalidPipelineRunsFilterError"

    def __init__(self, message):
        super().__init__()
        self.message = check.str_param(message, "message")


class GrapheneRunGroupNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "RunGroupNotFoundError"

    run_id = graphene.NonNull(graphene.String)

    def __init__(self, run_id):
        super().__init__()
        self.run_id = check.str_param(run_id, "run_id")
        self.message = f"Run group of run {run_id} could not be found."


class GraphenePresetNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PresetNotFoundError"

    preset = graphene.NonNull(graphene.String)

    def __init__(self, preset, selector):
        super().__init__()
        self.preset = check.str_param(preset, "preset")
        self.message = f"Preset {preset} not found in pipeline {selector.pipeline_name}."


class GrapheneConflictingExecutionParamsError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ConflictingExecutionParamsError"

    def __init__(self, conflicting_param):
        super().__init__()
        self.message = (
            f"Invalid ExecutionParams. Cannot define {conflicting_param} when using a preset."
        )


class GrapheneModeNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ModeNotFoundError"

    mode = graphene.NonNull(graphene.String)

    def __init__(self, mode, selector):
        super().__init__()
        self.mode = check.str_param(mode, "mode")
        self.message = f"Mode {mode} not found in pipeline {selector.pipeline_name}."


class GrapheneInvalidStepError(graphene.ObjectType):
    invalid_step_key = graphene.NonNull(graphene.String)

    class Meta:
        name = "InvalidStepError"


class GrapheneInvalidOutputError(graphene.ObjectType):
    step_key = graphene.NonNull(graphene.String)
    invalid_output_name = graphene.NonNull(graphene.String)

    class Meta:
        name = "InvalidOutputError"


class GraphenePipelineRunConflict(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PipelineRunConflict"


create_execution_params_error_types = (
    GraphenePresetNotFoundError,
    GrapheneConflictingExecutionParamsError,
)


class GrapheneDagsterTypeNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "DagsterTypeNotFoundError"

    dagster_type_name = graphene.NonNull(graphene.String)


class GrapheneScheduleNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "ScheduleNotFoundError"

    schedule_name = graphene.NonNull(graphene.String)

    def __init__(self, schedule_name):
        super().__init__()
        self.schedule_name = check.str_param(schedule_name, "schedule_name")
        self.message = f"Schedule {self.schedule_name} could not be found."


class GrapheneSensorNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "SensorNotFoundError"

    sensor_name = graphene.NonNull(graphene.String)

    def __init__(self, sensor_name):
        super().__init__()
        self.name = check.str_param(sensor_name, "sensor_name")
        self.message = f"Could not find `{sensor_name}` in the currently loaded repository."


class GraphenePartitionSetNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "PartitionSetNotFoundError"

    partition_set_name = graphene.NonNull(graphene.String)

    def __init__(self, partition_set_name):
        super().__init__()
        self.partition_set_name = check.str_param(partition_set_name, "partition_set_name")
        self.message = f"Partition set {self.partition_set_name} could not be found."


class GrapheneRepositoryNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "RepositoryNotFoundError"

    repository_name = graphene.NonNull(graphene.String)
    repository_location_name = graphene.NonNull(graphene.String)

    def __init__(self, repository_location_name, repository_name):
        super().__init__()
        self.repository_name = check.str_param(repository_name, "repository_name")
        self.repository_location_name = check.str_param(
            repository_location_name, "repository_location_name"
        )
        self.message = f"Could not find Repository {repository_location_name}.{repository_name}"


class GrapheneAssetNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "AssetNotFoundError"

    def __init__(self, asset_key):
        super().__init__()
        self.asset_key = check.inst_param(asset_key, "asset_key", AssetKey)
        self.message = f"Asset key {asset_key.to_string()} not found."


types = [
    GrapheneAssetNotFoundError,
    GrapheneConflictingExecutionParamsError,
    GrapheneDagsterTypeNotFoundError,
    GrapheneError,
    GrapheneInvalidOutputError,
    GrapheneInvalidPipelineRunsFilterError,
    GrapheneInvalidStepError,
    GrapheneModeNotFoundError,
    GraphenePartitionSetNotFoundError,
    GraphenePipelineNotFoundError,
    GraphenePipelineRunConflict,
    GraphenePipelineRunNotFoundError,
    GraphenePipelineSnapshotNotFoundError,
    GraphenePresetNotFoundError,
    GraphenePythonError,
    GrapheneReloadNotSupported,
    GrapheneRepositoryLocationNotFound,
    GrapheneRepositoryNotFoundError,
    GrapheneRunGroupNotFoundError,
    GrapheneScheduleNotFoundError,
    GrapheneSchedulerNotDefinedError,
    GrapheneSensorNotFoundError,
]



import graphene
from dagster import check
from graphene.types.generic import GenericScalar

from ..implementation.fetch_runs import get_runs, get_runs_count
from .errors import (
    GrapheneInvalidPipelineRunsFilterError,
    GraphenePythonError,
    GrapheneRunGroupNotFoundError,
)
from .util import non_null_list


class GrapheneStepEventStatus(graphene.Enum):
    SKIPPED = "SKIPPED"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    class Meta:
        name = "StepEventStatus"


class GrapheneLaunchPipelineRunSuccess(graphene.ObjectType):
    run = graphene.Field(
        graphene.NonNull("dagster_graphql.schema.pipelines.pipeline.GraphenePipelineRun")
    )

    class Meta:
        name = "LaunchPipelineRunSuccess"


class GrapheneRunGroup(graphene.ObjectType):
    rootRunId = graphene.NonNull(graphene.String)
    runs = graphene.List("dagster_graphql.schema.pipelines.pipeline.GraphenePipelineRun")

    class Meta:
        name = "RunGroup"

    def __init__(self, root_run_id, runs):
        from .pipelines.pipeline import GraphenePipelineRun

        check.str_param(root_run_id, "root_run_id")
        check.list_param(runs, "runs", GraphenePipelineRun)

        super().__init__(rootRunId=root_run_id, runs=runs)


class GrapheneRunGroups(graphene.ObjectType):
    results = non_null_list(GrapheneRunGroup)

    class Meta:
        name = "RunGroups"


launch_pipeline_run_result_types = (GrapheneLaunchPipelineRunSuccess,)


class GrapheneLaunchPipelineExecutionResult(graphene.Union):
    class Meta:
        from .backfill import pipeline_execution_error_types

        types = launch_pipeline_run_result_types + pipeline_execution_error_types

        name = "LaunchPipelineExecutionResult"


class GrapheneLaunchPipelineReexecutionResult(graphene.Union):
    class Meta:
        from .backfill import pipeline_execution_error_types

        types = launch_pipeline_run_result_types + pipeline_execution_error_types

        name = "LaunchPipelineReexecutionResult"


class GraphenePipelineRuns(graphene.ObjectType):
    results = non_null_list("dagster_graphql.schema.pipelines.pipeline.GraphenePipelineRun")
    count = graphene.Int()

    class Meta:
        name = "PipelineRuns"

    def __init__(self, filters, cursor, limit):
        super().__init__()

        self._filters = filters
        self._cursor = cursor
        self._limit = limit

    def resolve_results(self, graphene_info):
        return get_runs(graphene_info, self._filters, self._cursor, self._limit)

    def resolve_count(self, graphene_info):
        return get_runs_count(graphene_info, self._filters)


class GraphenePipelineRunsOrError(graphene.Union):
    class Meta:
        types = (GraphenePipelineRuns, GrapheneInvalidPipelineRunsFilterError, GraphenePythonError)
        name = "PipelineRunsOrError"


class GrapheneRunGroupOrError(graphene.Union):
    class Meta:
        types = (GrapheneRunGroup, GrapheneRunGroupNotFoundError, GraphenePythonError)
        name = "RunGroupOrError"


class GrapheneRunGroupsOrError(graphene.ObjectType):
    results = non_null_list(GrapheneRunGroup)

    class Meta:
        types = (GrapheneRunGroups, GraphenePythonError)
        name = "RunGroupsOrError"


class GrapheneRunConfigData(GenericScalar, graphene.Scalar):
    class Meta:
        description = """This type is used when passing in a configuration object
        for pipeline configuration. This is any-typed in the GraphQL type system,
        but must conform to the constraints of the dagster config type system"""
        name = "RunConfigData"


types = [
    GrapheneLaunchPipelineExecutionResult,
    GrapheneLaunchPipelineReexecutionResult,
    GrapheneLaunchPipelineRunSuccess,
    GraphenePipelineRuns,
    GraphenePipelineRunsOrError,
    GrapheneRunConfigData,
    GrapheneRunGroup,
    GrapheneRunGroupOrError,
    GrapheneRunGroups,
    GrapheneRunGroupsOrError,
]



from dagster.core.host_representation import RepositoryLocationManager
from dagster.core.host_representation.grpc_server_registry import ProcessGrpcServerRegistry
from dagster.daemon import get_default_daemon_logger
from dagster.daemon.sensor import execute_sensor_iteration
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    infer_job_selector,
    infer_repository_selector,
)

from .graphql_context_test_suite import (  # get_dict_recon_repo,
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)

GET_JOB_QUERY = """
query JobQuery($jobSelector: JobSelector!) {
  jobStateOrError(jobSelector: $jobSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on JobState {
        id
        nextTick {
            timestamp
        }
    }
  }
}
"""


def _create_sensor_tick(instance):
    with ProcessGrpcServerRegistry() as grpc_server_registry:
        with RepositoryLocationManager(grpc_server_registry) as location_manager:
            list(
                execute_sensor_iteration(
                    instance, get_default_daemon_logger("SensorDaemon"), location_manager
                )
            )


class TestNextTickRepository(
    make_graphql_context_test_suite(
        context_variants=GraphQLContextVariant.all_readonly_variants(),
    )
):
    def test_schedule_next_tick(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        external_repository = graphql_context.get_repository_location(
            repository_selector["repositoryLocationName"]
        ).get_repository(repository_selector["repositoryName"])
        graphql_context.instance.reconcile_scheduler_state(external_repository)

        schedule_name = "no_config_pipeline_hourly_schedule"
        external_schedule = external_repository.get_external_schedule(schedule_name)
        job_selector = infer_job_selector(graphql_context, schedule_name)

        # need to be running in order to generate a future tick
        graphql_context.instance.start_schedule_and_update_storage_state(external_schedule)
        result = execute_dagster_graphql(
            graphql_context, GET_JOB_QUERY, variables={"jobSelector": job_selector}
        )

        assert result.data
        assert result.data["jobStateOrError"]["__typename"] == "JobState"
        next_tick = result.data["jobStateOrError"]["nextTick"]
        assert next_tick

    def test_sensor_next_tick(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        external_repository = graphql_context.get_repository_location(
            repository_selector["repositoryLocationName"]
        ).get_repository(repository_selector["repositoryName"])
        graphql_context.instance.reconcile_scheduler_state(external_repository)

        sensor_name = "always_no_config_sensor"
        external_sensor = external_repository.get_external_sensor(sensor_name)
        job_selector = infer_job_selector(graphql_context, sensor_name)

        # need to be running and create a sensor tick in the last 30 seconds in order to generate a
        # future tick
        graphql_context.instance.start_sensor(external_sensor)
        _create_sensor_tick(graphql_context.instance)

        result = execute_dagster_graphql(
            graphql_context, GET_JOB_QUERY, variables={"jobSelector": job_selector}
        )

        assert result.data
        assert result.data["jobStateOrError"]["__typename"] == "JobState"
        next_tick = result.data["jobStateOrError"]["nextTick"]
        assert next_tick



import yaml
from dagster import check
from dagster.core.host_representation import (
    ExternalPartitionConfigData,
    ExternalPartitionExecutionErrorData,
    ExternalPartitionNamesData,
    ExternalPartitionSet,
    ExternalPartitionTagsData,
    RepositoryHandle,
    RepositorySelector,
)
from dagster.core.storage.pipeline_run import PipelineRunsFilter
from dagster.core.storage.tags import PARTITION_NAME_TAG, PARTITION_SET_TAG, TagType, get_tag_type
from graphql.execution.base import ResolveInfo

from .utils import capture_error


@capture_error
def get_partition_sets_or_error(graphene_info, repository_selector, pipeline_name):
    from ..schema.partition_sets import GraphenePartitionSet, GraphenePartitionSets

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(repository_selector, "repository_selector", RepositorySelector)
    check.str_param(pipeline_name, "pipeline_name")
    location = graphene_info.context.get_repository_location(repository_selector.location_name)
    repository = location.get_repository(repository_selector.repository_name)
    partition_sets = [
        partition_set
        for partition_set in repository.get_external_partition_sets()
        if partition_set.pipeline_name == pipeline_name
    ]

    return GraphenePartitionSets(
        results=[
            GraphenePartitionSet(
                external_repository_handle=repository.handle,
                external_partition_set=partition_set,
            )
            for partition_set in sorted(
                partition_sets,
                key=lambda partition_set: (
                    partition_set.pipeline_name,
                    partition_set.mode,
                    partition_set.name,
                ),
            )
        ]
    )


@capture_error
def get_partition_set(graphene_info, repository_selector, partition_set_name):
    from ..schema.partition_sets import GraphenePartitionSet, GraphenePartitionSetNotFoundError

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(repository_selector, "repository_selector", RepositorySelector)
    check.str_param(partition_set_name, "partition_set_name")
    location = graphene_info.context.get_repository_location(repository_selector.location_name)
    repository = location.get_repository(repository_selector.repository_name)
    partition_sets = repository.get_external_partition_sets()
    for partition_set in partition_sets:
        if partition_set.name == partition_set_name:
            return GraphenePartitionSet(
                external_repository_handle=repository.handle,
                external_partition_set=partition_set,
            )

    return GraphenePartitionSetNotFoundError(partition_set_name)


@capture_error
def get_partition_by_name(graphene_info, repository_handle, partition_set, partition_name):
    from ..schema.partition_sets import GraphenePartition

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.inst_param(partition_set, "partition_set", ExternalPartitionSet)
    check.str_param(partition_name, "partition_name")
    return GraphenePartition(
        external_repository_handle=repository_handle,
        external_partition_set=partition_set,
        partition_name=partition_name,
    )


def get_partition_config(graphene_info, repository_handle, partition_set_name, partition_name):
    from ..schema.errors import GraphenePythonError
    from ..schema.partition_sets import GraphenePartitionRunConfig

    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(partition_set_name, "partition_set_name")
    check.str_param(partition_name, "partition_name")

    result = graphene_info.context.get_external_partition_config(
        repository_handle,
        partition_set_name,
        partition_name,
    )

    if isinstance(result, ExternalPartitionConfigData):
        return GraphenePartitionRunConfig(
            yaml=yaml.safe_dump(result.run_config, default_flow_style=False)
        )
    else:
        return GraphenePythonError(result.error)


@capture_error
def get_partition_tags(graphene_info, repository_handle, partition_set_name, partition_name):
    from ..schema.errors import GraphenePythonError
    from ..schema.partition_sets import GraphenePartitionTags
    from ..schema.tags import GraphenePipelineTag

    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(partition_set_name, "partition_set_name")
    check.str_param(partition_name, "partition_name")

    result = graphene_info.context.get_external_partition_tags(
        repository_handle, partition_set_name, partition_name
    )

    if isinstance(result, ExternalPartitionTagsData):
        return GraphenePartitionTags(
            results=[
                GraphenePipelineTag(key=key, value=value)
                for key, value in result.tags.items()
                if get_tag_type(key) != TagType.HIDDEN
            ]
        )
    else:
        return GraphenePythonError(result.error)


@capture_error
def get_partitions(
    graphene_info, repository_handle, partition_set, cursor=None, limit=None, reverse=False
):
    from ..schema.errors import GraphenePythonError
    from ..schema.partition_sets import GraphenePartition, GraphenePartitions

    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.inst_param(partition_set, "partition_set", ExternalPartitionSet)
    result = graphene_info.context.get_external_partition_names(
        repository_handle, partition_set.name
    )

    if isinstance(result, ExternalPartitionNamesData):
        partition_names = _apply_cursor_limit_reverse(
            result.partition_names, cursor, limit, reverse
        )

        return GraphenePartitions(
            results=[
                GraphenePartition(
                    external_partition_set=partition_set,
                    external_repository_handle=repository_handle,
                    partition_name=partition_name,
                )
                for partition_name in partition_names
            ]
        )

    else:
        assert isinstance(result, ExternalPartitionExecutionErrorData)
        return GraphenePythonError(result.error)


def _apply_cursor_limit_reverse(items, cursor, limit, reverse):
    start = 0
    end = len(items)
    index = 0

    if cursor:
        index = next((idx for (idx, item) in enumerate(items) if item == cursor), None)

        if reverse:
            end = index
        else:
            start = index + 1

    if limit:
        if reverse:
            start = end - limit
        else:
            end = start + limit

    return items[max(start, 0) : end]


@capture_error
def get_partition_set_partition_statuses(graphene_info, repository_handle, partition_set_name):
    from ..schema.partition_sets import GraphenePartitionStatus, GraphenePartitionStatuses

    check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
    check.str_param(partition_set_name, "partition_set_name")
    result = graphene_info.context.get_external_partition_names(
        repository_handle, partition_set_name
    )
    all_partition_set_runs = graphene_info.context.instance.get_runs(
        PipelineRunsFilter(tags={PARTITION_SET_TAG: partition_set_name})
    )
    runs_by_partition = {}
    for run in all_partition_set_runs:
        partition_name = run.tags.get(PARTITION_NAME_TAG)
        if not partition_name or partition_name in runs_by_partition:
            # all_partition_set_runs is in descending order by creation time, we should ignore
            # runs for the same partition if we've already considered the partition
            continue
        runs_by_partition[partition_name] = run

    return GraphenePartitionStatuses(
        results=[
            GraphenePartitionStatus(
                id=f"{partition_set_name}:{partition_name}",
                partitionName=partition_name,
                runStatus=runs_by_partition[partition_name].status
                if runs_by_partition.get(partition_name)
                else None,
            )
            for partition_name in result.partition_names
        ]
    )



import json
import os
import tempfile
import time
from contextlib import contextmanager

from click.testing import CliRunner
from dagster import seven
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.test_utils import instance_for_test_tempdir
from dagster.utils import file_relative_path
from dagster_graphql.cli import ui


@contextmanager
def dagster_cli_runner():
    with tempfile.TemporaryDirectory() as dagster_home_temp:
        with instance_for_test_tempdir(
            dagster_home_temp,
            overrides={
                "run_launcher": {
                    "module": "dagster.core.launcher.sync_in_memory_run_launcher",
                    "class": "SyncInMemoryRunLauncher",
                }
            },
        ):
            yield CliRunner(env={"DAGSTER_HOME": dagster_home_temp})


def test_basic_introspection():
    query = "{ __schema { types { name } } }"

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ["-w", workspace_path, "-t", query])
        assert result.exit_code == 0

        result_data = json.loads(result.output)
        assert result_data["data"]


def test_basic_repositories():
    query = "{ repositoriesOrError { ... on RepositoryConnection { nodes { name } } } }"

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ["-w", workspace_path, "-t", query])

        assert result.exit_code == 0

        result_data = json.loads(result.output)
        assert result_data["data"]["repositoriesOrError"]["nodes"]


def test_basic_repository_locations():
    query = "{ repositoryLocationsOrError { ... on RepositoryLocationConnection { nodes { ... on RepositoryLocation { __typename, name } ... on RepositoryLocationLoadFailure { __typename, name, error { message } } } } } }"

    workspace_path = file_relative_path(__file__, "./cli_test_error_workspace.yaml")

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ["-w", workspace_path, "-t", query])

        assert result.exit_code == 0

        result_data = json.loads(result.output)

        nodes = result_data["data"]["repositoryLocationsOrError"]["nodes"]
        assert len(nodes) == 2

        assert nodes[0]["__typename"] == "RepositoryLocation"
        assert nodes[0]["name"] == "test_cli_location"

        assert nodes[1]["__typename"] == "RepositoryLocationLoadFailure"
        assert nodes[1]["name"] == "test_cli_location_error"
        assert "No module named" in nodes[1]["error"]["message"]


def test_basic_variables():
    query = """
    query FooBar($pipelineName: String! $repositoryName: String! $repositoryLocationName: String!){
        pipelineOrError(params:{pipelineName: $pipelineName repositoryName: $repositoryName repositoryLocationName: $repositoryLocationName})
        { ... on Pipeline { name } }
    }
    """
    variables = '{"pipelineName": "math", "repositoryName": "test", "repositoryLocationName": "test_cli_location"}'
    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")

    with dagster_cli_runner() as runner:
        result = runner.invoke(ui, ["-w", workspace_path, "-v", variables, "-t", query])

        assert result.exit_code == 0

        result_data = json.loads(result.output)
        assert result_data["data"]["pipelineOrError"]["name"] == "math"


LAUNCH_PIPELINE_EXECUTION_QUERY = """
mutation ($executionParams: ExecutionParams!) {
    launchPipelineExecution(executionParams: $executionParams) {
        __typename
        ... on LaunchPipelineRunSuccess {
            run {
                runId
                pipeline { ...on PipelineReference { name } }
            }
        }
        ... on PipelineConfigValidationInvalid {
            pipelineName
            errors { message }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


def test_start_execution_text():
    variables = seven.json.dumps(
        {
            "executionParams": {
                "selector": {
                    "repositoryLocationName": "test_cli_location",
                    "repositoryName": "test",
                    "pipelineName": "math",
                },
                "runConfigData": {"solids": {"add_one": {"inputs": {"num": {"value": 123}}}}},
                "mode": "default",
            }
        }
    )

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")

    with dagster_cli_runner() as runner:
        result = runner.invoke(
            ui, ["-w", workspace_path, "-v", variables, "-t", LAUNCH_PIPELINE_EXECUTION_QUERY]
        )

        assert result.exit_code == 0

        try:
            result_data = json.loads(result.output.strip("\n").split("\n")[-1])
            assert (
                result_data["data"]["launchPipelineExecution"]["__typename"]
                == "LaunchPipelineRunSuccess"
            )
        except Exception as e:
            raise Exception("Failed with {} Exception: {}".format(result.output, e))


def test_start_execution_file():
    variables = seven.json.dumps(
        {
            "executionParams": {
                "selector": {
                    "pipelineName": "math",
                    "repositoryLocationName": "test_cli_location",
                    "repositoryName": "test",
                },
                "runConfigData": {"solids": {"add_one": {"inputs": {"num": {"value": 123}}}}},
                "mode": "default",
            }
        }
    )

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")
    with dagster_cli_runner() as runner:
        result = runner.invoke(
            ui,
            [
                "-w",
                workspace_path,
                "-v",
                variables,
                "--file",
                file_relative_path(__file__, "./execute.graphql"),
            ],
        )

        assert result.exit_code == 0
        result_data = json.loads(result.output.strip("\n").split("\n")[-1])
        assert (
            result_data["data"]["launchPipelineExecution"]["__typename"]
            == "LaunchPipelineRunSuccess"
        )


def test_start_execution_save_output():
    """
    Test that the --output flag saves the GraphQL response to the specified file
    """

    variables = seven.json.dumps(
        {
            "executionParams": {
                "selector": {
                    "repositoryLocationName": "test_cli_location",
                    "repositoryName": "test",
                    "pipelineName": "math",
                },
                "runConfigData": {"solids": {"add_one": {"inputs": {"num": {"value": 123}}}}},
                "mode": "default",
            }
        }
    )

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")

    with dagster_cli_runner() as runner:
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = os.path.join(temp_dir, "output_file")

            result = runner.invoke(
                ui,
                [
                    "-w",
                    workspace_path,
                    "-v",
                    variables,
                    "--file",
                    file_relative_path(__file__, "./execute.graphql"),
                    "--output",
                    file_name,
                ],
            )

            assert result.exit_code == 0

            assert os.path.isfile(file_name)
            with open(file_name, "r") as f:
                lines = f.readlines()
                result_data = json.loads(lines[-1])
                assert (
                    result_data["data"]["launchPipelineExecution"]["__typename"]
                    == "LaunchPipelineRunSuccess"
                )


def test_start_execution_predefined():
    variables = seven.json.dumps(
        {
            "executionParams": {
                "selector": {
                    "repositoryLocationName": "test_cli_location",
                    "repositoryName": "test",
                    "pipelineName": "math",
                },
                "runConfigData": {"solids": {"add_one": {"inputs": {"num": {"value": 123}}}}},
                "mode": "default",
            }
        }
    )

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")

    with dagster_cli_runner() as runner:
        result = runner.invoke(
            ui, ["-w", workspace_path, "-v", variables, "-p", "launchPipelineExecution"]
        )
        assert result.exit_code == 0
        result_data = json.loads(result.output.strip("\n").split("\n")[-1])
        if not result_data.get("data"):
            raise Exception(result_data)
        assert (
            result_data["data"]["launchPipelineExecution"]["__typename"]
            == "LaunchPipelineRunSuccess"
        )


def test_logs_in_start_execution_predefined():
    variables = seven.json.dumps(
        {
            "executionParams": {
                "selector": {
                    "repositoryLocationName": "test_cli_location",
                    "repositoryName": "test",
                    "pipelineName": "math",
                },
                "runConfigData": {"solids": {"add_one": {"inputs": {"num": {"value": 123}}}}},
                "mode": "default",
            }
        }
    )

    workspace_path = file_relative_path(__file__, "./cli_test_workspace.yaml")
    with tempfile.TemporaryDirectory() as temp_dir:
        with instance_for_test_tempdir(
            temp_dir,
            overrides={
                "run_launcher": {
                    "module": "dagster.core.launcher.sync_in_memory_run_launcher",
                    "class": "SyncInMemoryRunLauncher",
                }
            },
        ) as instance:
            runner = CliRunner(env={"DAGSTER_HOME": temp_dir})
            result = runner.invoke(
                ui, ["-w", workspace_path, "-v", variables, "-p", "launchPipelineExecution"]
            )
            assert result.exit_code == 0
            result_data = json.loads(result.output.strip("\n").split("\n")[-1])
            assert (
                result_data["data"]["launchPipelineExecution"]["__typename"]
                == "LaunchPipelineRunSuccess"
            )
            run_id = result_data["data"]["launchPipelineExecution"]["run"]["runId"]

            # allow FS events to flush
            retries = 5
            while retries != 0 and not _is_done(instance, run_id):
                time.sleep(0.333)
                retries -= 1

            # assert that the watching run storage captured the run correctly from the other process
            run = instance.get_run_by_id(run_id)

            assert run.status == PipelineRunStatus.SUCCESS


def _is_done(instance, run_id):
    return instance.has_run(run_id) and instance.get_run_by_id(run_id).is_finished
